from dune.perftool.pdelab.argument import (name_coefficientcontainer,
                                           pymbolic_coefficient,
                                           )
from dune.perftool.generation import (backend,
                                      domain,
                                      function_mangler,
                                      get_counter,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      instruction,
                                      silenced_warning,
                                      temporary_variable,
                                      )
from dune.perftool.loopy.buffer import (get_buffer_temporary,
                                        initialize_buffer,
                                        switch_base_storage,
                                        )
from dune.perftool.pdelab.spaces import name_lfs
from dune.perftool.sumfact.amatrix import (AMatrix,
                                           quadrature_points_per_direction,
                                           basis_functions_per_direction,
                                           name_theta,
                                           name_theta_transposed,
                                           )
from dune.perftool.loopy.stages import stage_insn
from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )
from dune.perftool.sumfact.quadrature import quadrature_inames
from loopy import Reduction, GlobalArg
from loopy.symbolic import FunctionIdentifier

from pytools import product


class AccumulationArg(GlobalArg):
    allowed_extra_kwargs = GlobalArg.allowed_extra_kwargs + ["lfs"]


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


def setup_theta(element, restriction, component, a_matrices):
    number_basis = product(mat.cols for mat in a_matrices)
    shape = (number_basis,)
    inp = get_buffer_temporary("buffer",
                               shape=shape)
    silenced_warning('read_no_write({})'.format(inp))

    # Write initial coefficients into buffer
    basisiname = sumfact_iname(number_basis, "basis")
    lfs = name_lfs(element, restriction, component)
    container = name_coefficientcontainer(restriction)
    coeff = pymbolic_coefficient(container, lfs, basisiname)
    assignee = Subscript(Variable(inp), (Variable(basisiname),))
    return instruction(assignee=assignee,
                       expression=coeff,
                       depends_on=frozenset({stage_insn(0)}),
                       )


def pymbolic_trialfunction(element, restriction, component):
    theta = name_theta()
    rows = quadrature_points_per_direction()
    cols = basis_functions_per_direction()
    a_matrix = AMatrix(theta, rows, cols)
    a_matrices = (a_matrix, a_matrix)

    # Do stage 1
    initialize_buffer("buffer",
                      base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                      num=2
                      )

    insn_dep = setup_theta(element, restriction, component, a_matrices)
    var = sum_factorization_kernel(a_matrices, "buffer", 0, frozenset({insn_dep}))

    return Subscript(Variable(var), tuple(Variable(i) for i in quadrature_inames()))


@backend(interface="accum_insn", name="sumfact")
def generate_accumulation_instruction(visitor, accterm, measure, subdomain_id):
    pymbolic_expr = visitor(accterm.term)

    if pymbolic_expr == 0:
        return

    # Get geometric dimension
    formdata = get_global_context_value('formdata')
    dim = formdata.geometric_dimension

    # Get a matrix that holds the result of the pymbolic_expr at all quadrature points
    name = "blubb"
    temporary_variable(name, shape=(quadrature_points_per_direction(),) * dim, managed=True)
    instruction(assignee=Subscript(Variable(name), tuple(Variable(i) for i in quadrature_inames())),
                expression=pymbolic_expr,
                forced_iname_deps=frozenset(quadrature_inames()),
                forced_iname_deps_is_final=True,
                depends_on=frozenset({stage_insn(1)}),
                )

    theta_transposed = name_theta_transposed()
    rows = basis_functions_per_direction()
    cols = quadrature_points_per_direction()
    a_matrix = AMatrix(theta_transposed, rows, cols)
    a_matrices = (a_matrix, a_matrix)

    # Do stage 1
    initialize_buffer("reffub",
                      base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                      num=2
                      )

    result = sum_factorization_kernel(a_matrices, "reffub", 2)

    # Now write all this into the correct residual
    lfs = name_lfs(accterm.argument.argexpr.ufl_element(),
                   accterm.argument.restriction,
                   accterm.argument.component,
                   )
    inames = tuple(sumfact_iname(mat.rows, 'accum') for mat in a_matrices)
    globalarg("r",
              shape=(basis_functions_per_direction(),) * dim,
              argtype=AccumulationArg,
              lfs=lfs,
              )

    instruction(expression=Subscript(Variable(result), tuple(Variable(i) for i in inames)),
                assignee=Subscript(Variable('r'), tuple(Variable(i) for i in inames)),
                forced_iname_deps=frozenset(inames),
                forced_iname_deps_is_final=True,
                depends_on=frozenset({stage_insn(3)}),
                )


def sum_factorization_kernel(a_matrices, buffer, stage, insn_dep=frozenset({})):
    """
    Calculate a sum factorization matrix product.

    Y = A_{d-1}*...*A_0*X

    where X is the input matrix and Y is the output variable.

    Arguments:
    ----------
    a_matrices: An iterable of AMatrix instances
        The list of tensors to be applied to the input.
        Order of application is from 0 up.
    buffer: A string identifying the flip flop buffer in use
        for intermediate results. The memory is expected to be
        pre-initialized with the input.
    insn_dep: an instruction ID that the first issued instruction
        should depend upon. All following ones will depend on each
        other.
    """
    for l, a_matrix in enumerate(a_matrices):
        # Get a temporary that interprets the base storage of the input
        # as a column-major matrix. In later iteration of the amatrix loop
        # this reinterprets the output of the previous iteration.
        inp_shape = (a_matrix.cols, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))
        inp = get_buffer_temporary(buffer,
                                   shape=inp_shape,
                                   dim_tags="f,f")

        # The input temporary will only be read from, so we need to silence the loopy warning
        silenced_warning('read_no_write({})'.format(inp))

        switch_base_storage(buffer)

        # Get a temporary that interprets the base storage of the output
        # as row-major matrix
        out_shape = (a_matrix.rows, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))
        out = get_buffer_temporary(buffer,
                                   shape=out_shape,
                                   dim_tags="c,c")

        # Get the inames needed for one matrix-matrix multiplication
        i = sumfact_iname(out_shape[0], "row")
        j = sumfact_iname(out_shape[1], "col")
        k = sumfact_iname(a_matrix.cols, "red")

        # Construct the matrix-matrix-multiplication expression a_ik*in_kj
        from dune.perftool.sumfact.amatrix import ColMajorAccess
        prod = Product((Call(ColMajorAccess(a_matrix.a_matrix), (Variable(i), Variable(k))),
                        Subscript(Variable(inp), (Variable(k), Variable(j)))
                        ))

        # Issue the reduction instruction that implements the multiplication
        # at the same time store the instruction ID for the next instruction to depend on
        insn_dep = frozenset({instruction(assignee=Subscript(Variable(out), (Variable(i), Variable(j))),
                                          expression=Reduction("sum", k, prod),
                                          forced_iname_deps=frozenset({i, j}),
                                          forced_iname_deps_is_final=True,
                                          depends_on=insn_dep.union(frozenset({stage_insn(stage)})),
                                          )
                              })

    return out
